import http.server
import socketserver
import subprocess
import os
import signal
import urllib.parse
import re

PORT = 8014
HTML_TEMPLATE_PATH = 'taskapp/taskmgr.html'  # Path to your HTML template
UPLOAD_DIR = 'Apps'  # Directory where zip files will be uploaded and unzipped
app_ports = {}  # Store the mapping of PID to port

class TaskManagerHandler(http.server.BaseHTTPRequestHandler):

    def do_GET(self):
        if self.path.startswith('/action'):
            # Process action (start/stop)
            parsed_path = urllib.parse.urlparse(self.path)
            params = urllib.parse.parse_qs(parsed_path.query)
            action = params.get('action', [''])[0]
            app_name = params.get('app_name', [''])[0]
            pid = params.get('pid', [''])[0]
            if action == 'start' and app_name:
                # Start the app
                self.start_app(app_name)
            elif action == 'stop' and pid:
                # Stop the app
                self.stop_app(pid)
            # Redirect back to home page
            self.send_response(303)
            self.send_header('Location', '/')
            self.end_headers()
        else:
            # Serve the main page using the HTML template
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            html = self.generate_html()
            self.wfile.write(html.encode())

    def do_POST(self):
        if self.path == '/upload':
            self.handle_file_upload()

    def handle_file_upload(self):
        # Ensure the upload directory exists
        if not os.path.exists(UPLOAD_DIR):
            os.makedirs(UPLOAD_DIR)

        # Handle zip file upload manually
        content_length = int(self.headers['Content-Length'])
        boundary = self.headers['Content-Type'].split('=')[1].encode()
        file_data = self.rfile.read(content_length)

        # Split the multipart form-data
        fields = file_data.split(boundary)

        for field in fields:
            if b'filename=' in field:
                # Extract filename
                filename = re.findall(b'filename="([^"]+)"', field)[0].decode()
                file_content = field.split(b'\r\n\r\n')[1].split(b'\r\n')[0]
                
                if filename.endswith('.zip'):
                    zip_path = os.path.join(UPLOAD_DIR, filename)
                    
                    # Log upload details
                    print(f"Saving uploaded file to: {zip_path}")
                    print(f"File size: {len(file_content)} bytes")
                    print(f"First 100 bytes: {file_content[:100]}")

                    # Write the file to the directory
                    with open(zip_path, 'wb') as f:
                        f.write(file_content)

                    # Use the system unzip command instead of Python's zipfile module
                    self.unzip_file(zip_path)

                # Redirect to home after upload
                self.send_response(303)
                self.send_header('Location', '/')
                self.end_headers()

    def unzip_file(self, zip_path):
        # Use the system 'unzip' command to extract the file
        try:
            unzip_command = ['unzip', '-o', zip_path, '-d', UPLOAD_DIR]
            result = subprocess.run(unzip_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            if result.returncode == 0:
                print(f"Successfully extracted: {zip_path}")
                os.remove(zip_path)  # Optionally delete the zip file after extraction
            else:
                print(f"Failed to extract {zip_path}. Error: {result.stderr.decode()}")
        except Exception as e:
            print(f"Exception during extraction: {e}")

    def start_app(self, app_name):
        # Start the app using subprocess and capture the port
        try:
            current_dir = os.getcwd()
            print(f"Current working directory: {current_dir}")
            
            if not os.path.isabs(app_name):
                app_name = os.path.join(UPLOAD_DIR, app_name)

            print(f"Attempting to start app: {app_name}")

            python_path = '/accounts/1000/appdata/com.update.Term49.gYABgHeOVx7TI8ajV86LufpS8jM/data/py3.12/python3_ntoarmv7-qnx-static/tools/python3/python3'
            
            if not os.path.exists(python_path):
                print(f"Python interpreter not found at {python_path}, falling back to 'python3'")
                python_path = 'python3'  # Fallback to system python if specific one doesn't exist

            # Add -u to get ubuffered output for capturing port
            command = [python_path, '-u', app_name]

            print(f"Running command: {' '.join(command)}")

            process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

            # Capture the port from the output
            while True:
                output_line = process.stdout.readline().strip()
                if not output_line:
                    break
                print(f"App output: {output_line}")

                # Check if the line contains a port number (adjust this regex based on your output)
                match = re.search(r'(\d{4,5})', output_line)
                if match:
                    port = match.group(1)
                    app_ports[process.pid] = port
                    print(f"App started on port {port}")
                    break

            print(f'Successfully started app {app_name} with PID {process.pid}')
        except Exception as e:
            print(f'Error starting {app_name}: {e}')

    def stop_app(self, pid):
        try:
            os.kill(int(pid), signal.SIGTERM)
            print(f'Stopped process {pid}')
            if int(pid) in app_ports:
                del app_ports[int(pid)]  # Remove the port entry for the stopped process
        except Exception as e:
            print(f'Error stopping process {pid}: {e}')

    def generate_html(self):
        # Load HTML template
        try:
            with open(HTML_TEMPLATE_PATH, 'r') as file:
                html = file.read()

            # Insert running processes into HTML
            processes_html = self.generate_processes_html()
            html = html.replace('<!-- running_processes -->', processes_html)

            # Insert available apps into HTML
            apps_html = self.generate_apps_html()
            html = html.replace('<!-- available_apps -->', apps_html)

            return html
        except Exception as e:
            print(f'Error loading HTML template: {e}')
            return "<html><body><h1>Error loading HTML template</h1></body></html>"

    def generate_processes_html(self):
        # Generate HTML table rows for running Python apps
        processes = self.get_python_processes()
        rows = ''
        for pid, cmd in processes:
            # Check if this process has an associated port
            launch_link = ''
            if int(pid) in app_ports:
                port = app_ports[int(pid)]
                launch_link = f' - <a href="http://127.0.0.1:{port}" target="_blank">Launch UI</a>'
            rows += f'<tr><td>{pid}</td><td class="scrollable-command">{cmd}</td><td><a href="/action?action=stop&pid={pid}">Stop</a>{launch_link}</td></tr>'
        return rows

    def generate_apps_html(self):
        # Generate HTML for available apps to start
        apps = self.get_available_apps()
        app_list = ''
        for app in apps:
            app_list += f'<p>{app} - <a href="/action?action=start&app_name={urllib.parse.quote(app)}">Start</a></p>'
        return app_list

    def get_python_processes(self):
        # Get list of running python processes using pidin ar
        processes = []
        try:
            output = subprocess.check_output(['pidin', 'ar']).decode()
            lines = output.strip().split('\n')
            for line in lines:
                if 'python' in line:
                    parts = line.split()
                    pid = parts[0]
                    cmd = ' '.join(parts[1:])
                    if '/data/' in cmd:
                        index = cmd.find('/data/') + len('/data/')
                        cmd = cmd[index:]
                    processes.append((pid, cmd))
        except Exception as e:
            print(f'Error getting processes: {e}')
        return processes

    def get_available_apps(self):
        # Return a list of available apps to start
        apps = self.scan_apps_directory()
        return apps

    def scan_apps_directory(self):
        # Scan the 'Apps2' directory for .py files
        apps = []
        try:
            for root, dirs, files in os.walk(UPLOAD_DIR):
                for file in files:
                    if file.endswith('.py'):
                        rel_path = os.path.relpath(os.path.join(root, file), UPLOAD_DIR)
                        apps.append(rel_path)
        except Exception as e:
            print(f'Error scanning apps directory: {e}')
        return apps

if __name__ == '__main__':
    with socketserver.TCPServer(("", PORT), TaskManagerHandler) as httpd:
        print(f"Serving on port {PORT}")
        httpd.serve_forever()
